#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* structures, protos and errors of the lib */
#include "libdtm.h"
#include "libdtm_gdhe.h"

static int	lastUpdateDate = 0; 

/**************************************************************
 *
 * Creation of a structure GDHE_MNT
 *
 **************************************************************/

GDHE_MNT*
dtm_createGdheMnt ( )

{
  GDHE_MNT*	gdhe_dtm;	/* Ce que l'on va créer */

  /* Simple allocation */
  gdhe_dtm = (GDHE_MNT*) malloc (sizeof (GDHE_MNT));
  if (gdhe_dtm == NULL) {
    dtmError (ERR_DTM_ALLOC_GDHE_MNT);
    return (NULL);
  }
  
  /* Initialisation à NULL des tous les champs de la GDHE_MNT */
  memset ((GDHE_MNT*) gdhe_dtm, 0, sizeof (GDHE_MNT));
  gdhe_dtm->points = NULL;

  return (gdhe_dtm);
}

/**************************************************************
 *
 * Params check for the dtm_initGdheMnt fct.
 *
 **************************************************************/

int 
dtm_checkGdheInitParams (double	scale_x,
			 double	scale_y,
			 double	scale_z,
			 int	nb_x,
			 int	nb_y,
			 int	stockage)

{
  if ((scale_x <= 0.0) || (scale_y <= 0.0) || (scale_z <= 0.0))
    return dtmError(ERR_DTM_SCALE);
  
  if ((nb_x <= 0) || (nb_y <= 0))
    return dtmError(ERR_DTM_SIZE);
  
  if ((stockage != GDHE_MNT_X_COLUMN) && (stockage != GDHE_MNT_X_ROW))
    return dtmError(ERR_DTM_TYPE_STOCKAGE);
  
  return (DTM_OK);
}

/**************************************************************
 *
 * Initialisation of a GDHE_MNT structure
 *
 **************************************************************/

int
dtm_initGdheMnt (GDHE_MNT*	gdhe_dtm,
	       double		scale_x,
	       double      	scale_y,
	       double      	scale_z,
	       int		nbLines,
	       int		nbCols,
	       double		x_origine,
	       double      	y_origine,
	       double      	z_origine,
	       int		stockage)
     
{
  /* Parameters validity check */
  if (dtm_checkGdheInitParams (scale_x, scale_y, scale_z,
				  nbLines, nbCols, stockage) != DTM_OK)
    return (-1);
  
  /* Is the 'gdhe_dtm' already filled ? */
  if (gdhe_dtm->points != NULL)
    return dtmError(ERR_DTM_GDHE_POINTS_ALLOUES);

  /* Fill the gdhe_dtm header infos */
  gdhe_dtm->hdr.x0 = x_origine;
  gdhe_dtm->hdr.y0 = y_origine;
  gdhe_dtm->hdr.z0 = z_origine;

  gdhe_dtm->hdr.dx = scale_x;
  gdhe_dtm->hdr.dy = scale_y;
  gdhe_dtm->hdr.dz = scale_z;

  gdhe_dtm->hdr.nrows = nbLines;
  gdhe_dtm->hdr.ncols = nbCols;

  switch (stockage) {
  case GDHE_MNT_X_COLUMN:
    gdhe_dtm->hdr.nrows = nbCols;
    gdhe_dtm->hdr.ncols = nbLines;
    break;
  case GDHE_MNT_X_ROW:
    gdhe_dtm->hdr.nrows = nbLines;
    gdhe_dtm->hdr.ncols = nbCols;
    break;
  default:
    return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
/*     break; */
/* Unreached */
  }

  gdhe_dtm->hdr.storageType = stockage;

  /* Allocation of the point array */
  gdhe_dtm->points = (GDHE_MNT_POINT*) malloc (nbLines * nbCols * sizeof (GDHE_MNT_POINT));

  if (gdhe_dtm->points == NULL)
    return dtmError (ERR_DTM_ALLOC_GDHE_MNT);

  /* Initialize all the points at '0' */
  memset ((GDHE_MNT_POINT*) gdhe_dtm->points, 0, 
	  nbLines * nbCols * sizeof (GDHE_MNT_POINT));
  
  return (DTM_OK);
}

/***************************************************************
 *
 * Destruction of a GDHE_MNT.
 *
 ***************************************************************/
  
void 
dtm_destroyGdheMnt (GDHE_MNT*	gdhe_dtm)
     
{
  free (gdhe_dtm->points);
  free (gdhe_dtm);

  return;
}

/**************************************************************
 *
 * Copy of a GDHE_MNT - To be done
 *
 **************************************************************/

GDHE_MNT*
dtm_copyGdheMnt (GDHE_MNT*	gdhe_dtm)
     
{
  if (gdhe_dtm == NULL)
    return (NULL);

  printf ("ERR_DTM_FCT_NON_IMPLEMENTEE\n");
  return (NULL);
}

/**************************************************************
 *
 * Printfs infos on a GDHE_MNT
 *
 **************************************************************/

int
dtm_infoGdheMnt (GDHE_MNT*	gdhe_dtm)
     
{
  printf ("NbLines = %d, NbCols = %d\n", gdhe_dtm->hdr.nrows, gdhe_dtm->hdr.ncols);
  printf ("(x, y, z) origine : (%f, %f, %f)\n", 
	  gdhe_dtm->hdr.x0, gdhe_dtm->hdr.y0, gdhe_dtm->hdr.z0);
  printf ("Echelles en (x, y, z) : (%f, %f, %f)\n", 
	  gdhe_dtm->hdr.dx, gdhe_dtm->hdr.dy, gdhe_dtm->hdr.dz);
  return (DTM_OK);
}

/**************************************************************
 *
 * Printfs infos on a cell of a GDHE_MNT
 *
 **************************************************************/

int
dtm_infoGdheMntCell (GDHE_MNT*	gdhe_dtm,
		    int		i,
		    int		j)
     
{
  GDHE_MNT_POINT*	cell;

  if ((i < 0) || (i >= gdhe_dtm->hdr.nrows) || (j < 0) || (j >= gdhe_dtm->hdr.ncols)) {
    printf ("Entre des valeurs 0 <= I < %d et 0 <= J < %d !\n", 
	    gdhe_dtm->hdr.nrows, gdhe_dtm->hdr.ncols);
    return dtmError(ERR_DTM_INDICES);
  }
  
  cell = gdhe_dtm->points + (i * gdhe_dtm->hdr.ncols) + j;

  printf ("Point (%d, %d) du GDHE_MNT : Altitude %d (soit %f m), type %d\n",
	  i, j, cell->z, (cell->z * gdhe_dtm->hdr.dz) + gdhe_dtm->hdr.z0,
	  cell->type);

  return (DTM_OK);
}

/**********************************************************************
 *
 * params check for the GDHE_MNT production fct (dtm_dtm2GdheMnt)
 *
 **********************************************************************/

int dtm_CheckDtm2GdheMntParams (DTM*	dtm,
				      double	zScale,
				      double	zOffset)
{
  if ((zScale <= 0) || (zOffset < 0.0))
    return dtmError (ERR_DTM_SCALE);
  
  return (DTM_OK);
}

/**********************************************************************
 *
 * Filling of a GDHE_MNT structure from a DTM.
 *
 **********************************************************************/

int
dtm_dtm2GdheMnt (DTM*		dtm,
		 double		zScale,
		 double		zOffset,
		 int		luminance,
		 GDHE_MNT*	gdhe_dtm)


{
  int		i, j;		/* To loop on the cells */
  double	zOrigin;	/* Z value for wich z = 0 in the gdhe_dtm */
  DTM_CELL*	cell;		/* A cell of the original 'DTM' */
  GDHE_MNT_POINT*	gdhe_cell; /* A cell of the 'GDHE_MNT' */
  GDHE_MNT_STORAGE_TYPE	storage;

  /* Params validity */
  if (dtm_CheckDtm2GdheMntParams (dtm, zScale, zOffset) != DTM_OK)
    return (-1);

  zOrigin = - zOffset;

  switch (dtm->stockage) {
  case X_COLONNES:
    storage = GDHE_MNT_X_COLUMN;
    break;
  case X_LIGNES:
    storage = GDHE_MNT_X_ROW;
    break;
  default:
    return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
/*     break; */
/* Unreached */
  }

  /* Initialisation of the gdhe_dtm */
  if (dtm_initGdheMnt (gdhe_dtm, dtm->scale_x, dtm->scale_y, zScale,
		       dtm->nblig, dtm->nbcol,
		       dtm->x_origine, dtm->y_origine, zOrigin,
		       storage) != DTM_OK)
    return (-1);

  /* If a geosedic reference is set, then copy it in the origin */
/*   if (dtm->geoReference == GEO_REFERENCE_SET) { */
/*     switch (dtm->geoOrientation) { */
/*     case GEO_EAST: */
/*       gdhe_dtm->hdr.y0 = dtm->geodesicOrigin.northing; */
/*       gdhe_dtm->hdr.x0 = dtm->geodesicOrigin.easting */
/* 	+ (dtm->scale_x * dtm->nbcol); */
/*       break; */
/*     default: */
/*       return dtmError (ERR_DTM_FCT_NON_IMPLEMENTEE); */
/*       break; unreached */
/*     } */
/*   } */

  /* computes dtm extremum values */
  dtm_minmax (dtm);
  
  /* Loop on the cells */
  cell = dtm->cells_tab;
  gdhe_cell = gdhe_dtm->points;
  for (i = 0; i < dtm->nblig; i++) {
    for (j = 0; j < dtm->nbcol; j++) {
      if (cell->state == EMPTY) {
	gdhe_cell->z = 0;
	gdhe_cell->type = 0;
	
	if (cell->may_be_visible == SEEN_BOTH)
	  gdhe_cell->type = 3;
      }
      else {
	if (cell->hstate == VERTICAL) 
	  gdhe_cell->z = (short) (((cell->current_info.z_max - zOrigin) / zScale) + 0.5);
	else
	  gdhe_cell->z = (short) (((cell->current_info.z_moyen - zOrigin) / zScale) + 0.5);
	switch (luminance) {
	case 1:
	  gdhe_cell->type = cell->current_info.luminance;
	  break;
	case 2:
	  gdhe_cell->type = 255 * (cell->current_info.luminance - dtm->colorMin) 
			    / (dtm->colorMax - dtm->colorMin);
	case 3:
	  gdhe_cell->type = 255 * (cell->current_info.z_moyen - dtm->z_min) 
			    / (dtm->z_max - dtm->z_min);
	  break;
	case 4:
	  gdhe_cell->type = 255 * (cell->current_info.nb_points - dtm->nbPtsMin) 
			    / (dtm->nbPtsMax - dtm->nbPtsMin);
	  break;
	case 5:
	  gdhe_cell->type = 255 * (cell->current_info.sigma_z - dtm->sigmaMin) 
			    / (dtm->sigmaMax - dtm->sigmaMin);
	  break;
	default:
	  if (cell->hstate == VERTICAL) 
	    gdhe_cell->type = 2;
	  else if (cell->mstate == DYNAMIC) 
	    gdhe_cell->type = 3;	
	  else
	    gdhe_cell->type = 1;	/* Pour pouvoir avoir une couleur dans gdhe */
	  break;
	}
      }
      cell++; 
      gdhe_cell++;
    }
  }
  
  /* Sets the last updated date to the current date of the DTM */
  lastUpdateDate = dtm->nb_im3d;

  return (DTM_OK);
}

/**********************************************************************
 *
 * Filling of a GDHE_MNT structure from a DTM, but considering only the last
 * modified cells of the DTM.
 *
 **********************************************************************/

int
dtm_lastDtm2GdheMnt (DTM*	dtm,
		     double	zScale,
		     double	zOffset,
		     int	luminance,
		     GDHE_MNT*	gdhe_dtm)


{
  int		i, j;		/* To loop on the cells */
  double	xOrigin;
  double	yOrigin;	/* Coordinates of the top-left corner of the
				   top-left sub-dtm in the global frame */
  double	zOrigin;	/* Z value for wich z = 0 in the gdhe_dtm */
  int		lastDate;	/* date of the last dtm update */
  int		iMin, iMax;
  int		jMin, jMax;	/* Bounds of the last dtm update */
  int		nblig_gdhe;
  int		nbcol_gdhe;	/* Size of the gdhe_dtm */
  DTM_CELL*	cell;		/* A cell of the original 'DTM' */
  GDHE_MNT_POINT*	gdhe_cell; /* A cell of the 'GDHE_MNT' */
  GDHE_MNT_STORAGE_TYPE	storage;


  /* Params validity */
  if (dtm_CheckDtm2GdheMntParams (dtm, zScale, zOffset) != DTM_OK)
    return (-1);

  /* Compute the dtm bounds of the last perception, in order to determine the
     size of the gdhe_dtm. Also, find the dtm min elevation in the last
     perception */ 
  lastDate = dtm->nb_im3d;
  iMin = dtm->nblig;
  iMax = 0;
  jMin = dtm->nbcol;
  jMax = 0;
  cell = dtm->cells_tab;
  for (i = 0; i < dtm->nblig; i++) {
    for (j = 0; j < dtm->nbcol; j++) {
      if ((cell->state == FILLED) && (cell->current_info.lastDate > lastUpdateDate)) {
	iMin = MIN (i, iMin);
	iMax = MAX (i, iMax);
	jMin = MIN (j, jMin);
	jMax = MAX (j, jMax);
      }
      cell++;
    }
  }
  nblig_gdhe = iMax - iMin + 1;
  nbcol_gdhe = jMax - jMin + 1;
  if ((nblig_gdhe <= 0) || (nbcol_gdhe <= 0)) {
    /* Vicious case where no pixels has been fused in the dtm (it happens,
       when it gets really dark) */
    return (ERR_DTM_NO_CELLS_UPDATED);
  }
  zOrigin = - zOffset;  

  switch (dtm->stockage) {
  case X_COLONNES:
    storage = GDHE_MNT_X_COLUMN;
    break;
  case X_LIGNES:
    storage = GDHE_MNT_X_ROW;
    break;
  default:
    return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
/*     break; */
/* Unreached */
  }

  /* Initialisation of the gdhe_dtm. */
  xOrigin = iMin * dtm->scale_x + dtm->x_origine;
  yOrigin = jMin * dtm->scale_y + dtm->y_origine;

  if (dtm_initGdheMnt (gdhe_dtm, dtm->scale_x, dtm->scale_y, zScale,
		       nblig_gdhe, nbcol_gdhe,
		       xOrigin, yOrigin, zOrigin,
		       storage) != DTM_OK)
    return (-1);
  
  /* Loop on the cells of the last perception */
  for (i = 0; i < nblig_gdhe; i++) {
    for (j = 0; j < nbcol_gdhe; j++) {
      cell = dtm->cells_tab + ((i + iMin) * dtm->nbcol) + (j + jMin);
      gdhe_cell = gdhe_dtm->points + (i * (gdhe_dtm->hdr).ncols) + j;
      if ((cell->state == FILLED) && (cell->current_info.lastDate >= lastUpdateDate)) {
	if (cell->hstate == VERTICAL) 
	  gdhe_cell->z = (short) (((cell->current_info.z_max - zOrigin) / zScale) + 0.5);
	else
	  gdhe_cell->z = (short) (((cell->current_info.z_moyen - zOrigin) / zScale) + 0.5);
	switch (luminance) {
	case 1:
	  gdhe_cell->type = cell->current_info.luminance;
	  break;
	case 2:
	  gdhe_cell->type = 255 * (cell->current_info.luminance - dtm->colorMin) 
	    / (dtm->colorMax - dtm->colorMin);
	case 3:
	  gdhe_cell->type = 255 * (cell->current_info.z_moyen - dtm->z_min) 
	    / (dtm->z_max - dtm->z_min);
	  break;
	case 4:
	  gdhe_cell->type = 255 * (cell->current_info.nb_points - dtm->nbPtsMin) 
	    / (dtm->nbPtsMax - dtm->nbPtsMin);
	  break;
	case 5:
	  gdhe_cell->type = 255 * (cell->current_info.sigma_z - dtm->sigmaMin) 
	    / (dtm->sigmaMax - dtm->sigmaMin);
	  break;
	default:
	  if (cell->hstate == VERTICAL) 
	    gdhe_cell->type = 2;
	  else if (cell->mstate == DYNAMIC) 
	    gdhe_cell->type = 3;	
	  else
	    gdhe_cell->type = 1;	/* Pour pouvoir avoir une couleur dans gdhe */
	  break;
	}
      }
      else {
	gdhe_cell->z = 0;
	gdhe_cell->type = 0;
      }
    }
  }
  
  lastUpdateDate = lastDate;

  return (DTM_OK);
}

/**********************************************************************
 *
 * Filling of a DTM structure from a GDHE_MNT. Just to debug.
 *
 **********************************************************************/

DTM*
dtm_gdheMnt2Dtm (GDHE_MNT*	gdhe_dtm)

{
  DTM*		dtm;		/* What is created here an returned */
  DTM_CELL*	cell;		/* A cell of the DTM */
  GDHE_MNT_POINT*	gdhe_cell; /* A cell of the GDHE_MNT */
  int		i, j;
  double	zOrigin;
  double	zScale;
  dtm_stockage	stockage;

  /* Creation of the DTM structure */
  dtm = create_dtm ();
  if (dtm == NULL)
    return (NULL);

  switch (gdhe_dtm->hdr.storageType) {
  case GDHE_MNT_X_COLUMN:
    stockage = X_COLONNES;
    break;
  case GDHE_MNT_X_ROW:
    stockage = X_LIGNES;
    break;
  default:
    dtmError(ERR_DTM_CAS_IMPOSSIBLE);
    return (NULL);
/*     break; */
/* Unreached */
  }

  if (init_dtm_cells (dtm, gdhe_dtm->hdr.dx, gdhe_dtm->hdr.dy, 
		      gdhe_dtm->hdr.nrows, gdhe_dtm->hdr.ncols, 
		      gdhe_dtm->hdr.x0, gdhe_dtm->hdr.y0, 
		      DTM_FIXED, stockage) != DTM_OK)
    return (NULL);
  
  zOrigin = gdhe_dtm->hdr.z0;
  zScale = gdhe_dtm->hdr.dz;

  cell = dtm->cells_tab;
  gdhe_cell = gdhe_dtm->points;
  for (i = 0; i < dtm->nblig; i++) {
    for (j = 0; j < dtm->nbcol; j++) {
      if ((gdhe_cell->z == 0) && (gdhe_cell->type == 0)) 
	cell->state = EMPTY;
      else {
	cell->state = FILLED;
	cell->current_info.z_moyen = (zScale * gdhe_cell->z) + zOrigin;
	cell->current_info.nb_points = 1;
	dtm->nb_cells++;
      }
      cell++; 
      gdhe_cell++;
   }
  }

  dtm->cells = CELLS_DETERMINEES;


  return (dtm);
}

/**********************************************************************
 *
 * Writing of a GDHE_MNT structure, so that GDHE can read it.
 *
 **********************************************************************/

int
dtm_writeGdheMnt (GDHE_MNT*	gdhe_dtm,
		  FILE*		fichier)

{
  int		nbCells;	/* Number of cells in the gdhe_dtm */

  if (fwrite (gdhe_dtm, sizeof (GDHE_MNT_HEADER), 1, fichier) != 1)
    return dtmError (ERR_DTM_WRITE_FILE);

  nbCells = gdhe_dtm->hdr.nrows * gdhe_dtm->hdr.ncols;

  if (fwrite (gdhe_dtm->points, sizeof (GDHE_MNT_POINT), nbCells, fichier) != nbCells)
    return dtmError (ERR_DTM_WRITE_FILE);

  return (DTM_OK);  
}

/**********************************************************************
 *
 * Reading of a GDHE_MNT structure
 *
 **********************************************************************/

GDHE_MNT*
dtm_readGdheMnt (char*		fileName)

{
  FILE*		fichier;
  GDHE_MNT*	gdhe_dtm;
  int		nbCells;	/* Number of cells in the gdhe_dtm */
  int		nbCellsRead;

  /* Création du GDHE_MNT */
  gdhe_dtm = dtm_createGdheMnt ();
  if (gdhe_dtm == NULL) {
    dtmError (ERR_DTM_ALLOC_GDHE_MNT);
    return (NULL);
  }
  
  fichier = fopen (fileName, "r");
  if (fichier == NULL) {
    dtmError(ERR_DTM_OPEN_FILE);
    return (NULL);
  }

  /* Read the header */
  if (fread (gdhe_dtm, sizeof (GDHE_MNT_HEADER), 1, fichier) != 1) {
    dtmError (ERR_DTM_READ_FILE);
    return (NULL);
  }

  /* Initialisation of the gdhe_dtm */
  if (dtm_initGdheMnt (gdhe_dtm, gdhe_dtm->hdr.dx, gdhe_dtm->hdr.dy, gdhe_dtm->hdr.dz,
		     gdhe_dtm->hdr.nrows, gdhe_dtm->hdr.ncols,
		     gdhe_dtm->hdr.x0, gdhe_dtm->hdr.y0, gdhe_dtm->hdr.z0,
		     gdhe_dtm->hdr.storageType) != DTM_OK) {
    dtmError (ERR_DTM_ALLOC_GDHE_MNT);
    return (NULL);
  }

  nbCells = gdhe_dtm->hdr.nrows * gdhe_dtm->hdr.ncols;

  nbCellsRead = fread (gdhe_dtm->points, sizeof (GDHE_MNT_POINT), nbCells, fichier);
  if (nbCellsRead != nbCells) {
    dtmError (ERR_DTM_READ_FILE);
    return (NULL);
  }
  
  fclose (fichier);

  return (gdhe_dtm);  
}

