/* Les includes génériques de C */
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <math.h>

/* Les structures, protos et erreurs de la lib */
#include "libdtm.h"

struct point {
	int i;
	int j;
};


struct extremum {
	struct point points[4];
};

struct real_point {
	double x;
	double y;
};

struct real_extremum {
	struct real_point points[4];
};

struct min_max {
	int min;
	int max;
};

struct tab_min_max {
	struct min_max* tab;
	size_t size;
};

struct dynamic_fuse_data {
	struct real_extremum ext;
	struct tab_min_max tab;
	dtm_dynamic_fuse_strategy strategy;
};

static void 
clear_dynamic_fuse_data(DTM* dtm, struct dynamic_fuse_data* data)
{
#ifdef APPLY_VISIBILITY
	size_t i;
	for (i = 0; i < data->tab.size; i++) {
		data->tab.tab[i].min = dtm->nblig;
		data->tab.tab[i].max = -1;
	}
#endif
}

/*
 * Modify center of the dtm so the robot is still at the center of the dtm It
 * will be used in the local dtm case Code is really copied from slideDtm, but
 * we don't need to explicitly slide it, because the dtm is reset at each
 * iteration
 */
static int
set_dtm_center(DTM * dtm, double x, double y)
{
  int	i_robot, j_robot;	/* Robot in integer coordinates */
  int	i0, j0;			/* Integer coordinates of the world reference
				   frame */ 
  int	nb_pix_metre_i;
  int	nb_pix_metre_j;		/* Number of dtm cells per meter */

  int   shift_i, shift_j;	/* Translation to apply on the DTM  */
  i0 = dtm->i_origine;
  j0 = dtm->j_origine;

  switch (dtm->stockage) {
  case X_LIGNES:
    nb_pix_metre_i = lround (1.0 / dtm->scale_x);
    nb_pix_metre_j = lround (1.0 / dtm->scale_y);
    i_robot = nb_pix_metre_i * x + i0;
    j_robot = nb_pix_metre_j * y + j0;
    break;
  case X_COLONNES:
    return dtmError (ERR_DTM_FCT_NON_IMPLEMENTEE);
    /*     break; */
    /* Unreached */
  default:
    /*     break; */
    /*     Unreached */
    return dtmError (ERR_DTM_CAS_IMPOSSIBLE);
  }

  shift_i = i_robot - (dtm->nblig / 2);
  shift_j = j_robot - (dtm->nbcol / 2);

  dtm->x_origine += shift_i * dtm->scale_x;
  dtm->y_origine += shift_j * dtm->scale_y;

  /* Détermination des coordonnées pixels de l'origine du repère global */
  switch (dtm->stockage) {
  case X_LIGNES:
    dtm->i_origine = (int) (floor (- dtm->x_origine / dtm->scale_x));
    dtm->j_origine = (int) (floor (- dtm->y_origine / dtm->scale_y));
    break;
  case X_COLONNES:
    dtm->j_origine = (int) (floor (- dtm->x_origine / dtm->scale_x));
    dtm->i_origine = (int) (floor (- dtm->y_origine / dtm->scale_y));
    break;
  default:
    return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
/*     break; */
/* Unreached */
  }

  return 0;
}

static void
__compute_extremum_visible_cell(
		T3D* sensorToMain,
		double u0, double v0, double alpha_u, double alpha_v,
		int u, int v,
		struct real_extremum* rext,
		int point)
{
	double tan_theta = (u - u0) / alpha_u;
	double theta = atan(tan_theta);
	double cos_theta = cos(theta);
	double tan_phi = cos_theta * ((v - v0)/alpha_v);
	double t;
	/* The coeficients of its parametric equation in the camera frame are
	   the following */  
	double alphaX = tan_theta;
	double alphaY = tan_phi / cos_theta;
	double alphaZ = 1;

	double betaX, betaY, betaZ;

	/* And in the DTM frame */
	betaX = alphaX * sensorToMain->matrix.matrix[0]
		+ alphaY * sensorToMain->matrix.matrix[1]
		+ alphaZ * sensorToMain->matrix.matrix[2];
	betaY = alphaX * sensorToMain->matrix.matrix[4]
		+ alphaY * sensorToMain->matrix.matrix[5]
		+ alphaZ * sensorToMain->matrix.matrix[6];
	betaZ = alphaX * sensorToMain->matrix.matrix[8]
		+ alphaY * sensorToMain->matrix.matrix[9]
		+ alphaZ * sensorToMain->matrix.matrix[10];

	/* Compute the coordinates of the line end in the world frame */
	t = sensorToMain->matrix.matrix[11] / betaZ;  
	rext->points[point].x = t  * -betaX + sensorToMain->matrix.matrix[3];
	rext->points[point].y = t  * -betaY + sensorToMain->matrix.matrix[7];
}

static void
_compute_extremum_visible_cell(
		T3D* sensorToMain, 
		double*	intrinsincParams,
		int		nbLines,
		int		nbColumns,
		int		reduction,
		struct  real_extremum* rext,
		int		first_point,
		int		second_point,
		double	max_depth,
		bool	right)
{
  double	u0, v0;
  double	alpha_u, alpha_v; /* Camera calibration parameters */
  
  t3dConvertTo(T3D_MATRIX, sensorToMain);
  
  /* Get the calibration params */
  v0 = *(intrinsincParams + 5) / reduction; 
  u0 = *(intrinsincParams + 2) / reduction;
  alpha_v = *(intrinsincParams + 4) / reduction;
  alpha_u = *(intrinsincParams) / reduction; 
  
  int v;
  
  if (right) { /* Right camera. Get the point corresponding to pixel
		  (0,nbLines) ("bottom-left" point of the image) */ 
    __compute_extremum_visible_cell(sensorToMain, u0, v0, alpha_u, alpha_v,
				    0, nbLines, rext, first_point);
    
    v = nbLines; /* Point corresponding to the pixel (0, v) that is not
		    further than max_depth ("top-left" point of the image) */
    __compute_extremum_visible_cell(sensorToMain, u0, v0, alpha_u, alpha_v,
				    0, v, rext, second_point);
    while (v > 0 && rext->points[second_point].x < max_depth) {
      __compute_extremum_visible_cell(sensorToMain, u0, v0, alpha_u, alpha_v,
				      0, v, rext, second_point);
      v--;
    }
    printf ("V found for the top-left point : %d\n", v);

  } else { /* Left camera. Get the point corresponding to pixel (0,nbLines)
	      ("bottom-right" point of the image) */ 
    __compute_extremum_visible_cell(sensorToMain, u0, v0, alpha_u, alpha_v,
				    nbColumns, nbLines, rext, first_point);

    v = nbLines; /* Point corresponding to the pixel (0, v) that is not
		    further than max_depth ("top-right" point of the image) */
    __compute_extremum_visible_cell(sensorToMain, u0, v0, alpha_u, alpha_v,
				    nbColumns, v, rext, second_point);
    while (v > 0 && rext->points[second_point].x < max_depth) {
      __compute_extremum_visible_cell(sensorToMain, u0, v0, alpha_u, alpha_v,
				      nbColumns, v, rext, second_point);
      v--;
    }
    printf ("V found for the top-right point : %d\n", v);
  }
}

static void 
bresenham(struct point p1, struct point p2, struct tab_min_max * tab)
{
	int	xdelta;			/* The change in x coordinates */
	int	ydelta;			/* The change in y coordinates */
	int	xstep;			/* The change to make in the x coordinate in
						 * each step */
	int	ystep;			/* The change to make in the y coordinate in
						 * each step */
	int	change;			/* The amount that the x or y coordinate has
						 * changed */
	int i,j;

	i = p1.i;
	j = p1.j;

	xdelta = p2.i  - p1.i;
	ydelta = p2.j - p1.j;

	if (xdelta < 0) {		/* The line will be drawn from right to left */
		xdelta = -xdelta;
		xstep = -1;
	} else			/* The line will be drawn from left to right */
		xstep = 1;
	if (ydelta < 0) {		/* The line will be drawn from bottom to top */
		ydelta = -ydelta;
		ystep = -1;
	} else			/* The line will be drawn from top to bottom */
		ystep = 1;

	if (xdelta > ydelta) {	/* x changes quicker than y */
		change = xdelta >> 1;	/* change set to twice the value of xdelta */
		while (i != p2.i) {	/* Draw until the terminating dot is reached */

			if (tab->tab[j].min > i) 
				tab->tab[j].min = i;

			if (tab->tab[j].max < i) 
				tab->tab[j].max = i;

			i += xstep;		/* Update x coordinate */
			change += ydelta;		/* Update change */
			if (change > xdelta)
				/*
				 * If change is large enough to update the y coordinate
				 */
			{
				j += ystep;		/* Update the y coordinate */
				change -= xdelta;	/* Reset change */
			}
		}
	} else {			/* y changes quicker than x */
		change = ydelta >> 1;	/* change set to twice the value of ydelta */
		while (j != p2.j) {		/* Draw until the terminating dot is reached */

			if (tab->tab[j].min > i) 
				tab->tab[j].min = i;

			if (tab->tab[j].max < i) 
				tab->tab[j].max = i;

			/* (i1, j1) is a point of the line. Compute the corresponding line
			   altitude */ 
			j += ystep;		/* Update y coordinate */
			change += xdelta;		/* Update change */
			if (change > ydelta)
				/* If change is large enough to update the x coordinate */
			{
				i += xstep;		/* Update the x coordinate */
				change -= ydelta;	/* Reset change */
			}
		}
	}
}

static void 
paint(DTM* dtm, struct tab_min_max *limits)
{
	size_t j;
	for (j= 0; j < limits->size; j++)
	{
		int i;
		int min = limits->tab[j].min;
		int max = limits->tab[j].max;
		if (max != -1) 
		{
			for (i = min; i < max; i++) 
			{
				DTM_CELL *cell = dtm_one_cell(dtm, i, j);
				cell->may_be_visible = SEEN_BOTH;
			}
		}
	}
}

static void
dtm_painting_visibility(DTM* dtm, struct extremum* ext, struct tab_min_max * limits)
{
	bresenham(ext->points[0], ext->points[1], limits);
	bresenham(ext->points[0], ext->points[2], limits);
	bresenham(ext->points[1], ext->points[3], limits);
	bresenham(ext->points[2], ext->points[3], limits);
	paint(dtm, limits);
}

static void
dtm_apply_visibility_pattern(DTM* dtm, 
			     double Tx, 
			     double Ty, 
			     double theta)
{
  int i;
  double cos_theta = cos(theta);
  double sin_theta = sin(theta);
  struct dynamic_fuse_data* data = dtm->data;
  struct real_extremum *ext = &data->ext;
  struct tab_min_max *limits = &data->tab;
  struct extremum xext;
  
  
  for (i = 0; i < 4; i++) {
    double x,y;
    double xxformed, yxformed;
    int i_dtm, j_dtm;
    DTM_CELL * cell;
    
    x = ext->points[i].x;
    y = ext->points[i].y;
    
    xxformed = Tx + x * cos_theta  - y * sin_theta;
    yxformed = Ty + x * sin_theta + y * cos_theta;
    
    i_dtm = (xxformed - dtm->x_origine) / dtm->scale_x;
    j_dtm = (yxformed - dtm->y_origine) / dtm->scale_y;
    
    if (i_dtm < 0 || i_dtm > dtm->nblig || j_dtm < 0 || j_dtm > dtm ->nbcol) {
      continue;
    }
    
    cell = dtm_one_cell(dtm, i_dtm, j_dtm);
    cell->may_be_visible = SEEN_BOTH;
    xext.points[i].i = i_dtm;
    xext.points[i].j = j_dtm;
  }
  
  dtm_painting_visibility(dtm, &xext, limits);
}

/*
 * Implémentation basique qui ne tient pas en compte la distance De plus, il
 * faudrait avoir un moyen de modifier facilement le seuil à partir de
 * l'appelant
 */
static inline void
evaluate_cell_hstate(DTM_CELL * cell, float distance, float threshold)
{
	(void) distance;
	if (cell->current_info.sigma_z > threshold) 
		cell->hstate = VERTICAL;
	else 
		cell->hstate = FLAT;
}

/*
 * Parcourt l'ensemble du DTM, met le dynamic state à STATIC, et évalue pour
 * toutes les nodes l'état hstate (verticale ou plat).
 */
static void
evaluate_hstate(DTM *mnt, T3D* RobotRed_2_Start, dtm_dynamic_fuse_strategy strategy)
{
  int			nblig_mnt;
  int			nbcol_mnt;
  int nb_pix_metre_i, nb_pix_metre_j;
  int i_robot, j_robot;
  int i0, j0;
  float sigma_mean = 0.0;
  int sigma_nb = 0;
  int		i, j;
  DTM_CELL* cell;
  double treshold;

  t3dConvertTo (T3D_EULER, RobotRed_2_Start);
  i0 = mnt->i_origine;
  j0 = mnt->j_origine;
  nb_pix_metre_i = lround (1.0 / mnt->scale_x);
  nb_pix_metre_j = lround (1.0 / mnt->scale_y);
  i_robot = nb_pix_metre_i * RobotRed_2_Start->euler.euler[3] + i0;
  j_robot = nb_pix_metre_j * RobotRed_2_Start->euler.euler[4] + j0;
  nblig_mnt = mnt->nblig;
  nbcol_mnt = mnt->nbcol;

  if (strategy.strat == MEAN_TRESHOLD) {
	  /* Get the bounds of the perceived area and the average of the standard
		 deviation */
	  cell = mnt->cells_tab;
	  for (i = 0; i < mnt->nblig; i++) {
		  for (j = 0; j < mnt->nbcol; j++) {
			  if (cell->state == FILLED) {
				  mnt->i_max = MAX (mnt->i_max, i);
				  mnt->i_min = MIN (mnt->i_min, i);
				  mnt->j_max = MAX (mnt->j_max, j);
				  mnt->j_min = MIN (mnt->j_min, j);

				  sigma_nb ++;
				  sigma_mean += cell->current_info.sigma_z;
			  }
			  cell++;
		  }
	  }

	  if (sigma_nb != 0)
		  sigma_mean /= sigma_nb;

	  for (i = mnt->i_min; i < mnt->i_max; i++) {
		  for (j = mnt->j_min; j < mnt->j_max; j++) {
			  cell = dtm_one_cell(mnt, i, j);
			  if (cell->state == FILLED) {
				  cell->mstate = STATIC;
				  evaluate_cell_hstate(cell, 0, strategy.treshold * sigma_mean);
			  }
		  }
	  }
  } else { /* FIXED_TRESHOLD */
	  cell = mnt->cells_tab;
	  for (i = 0; i < mnt->nblig; i++) {
		  for (j = 0; j < mnt->nbcol; j++) {
			  if (cell->state == FILLED) {
				  cell->mstate = STATIC;
				  evaluate_cell_hstate(cell, 0, strategy.treshold);
			  }
			  cell++;
		  }
	  }
  }
}

static void
dtm_merge_cell(DTM_CELL * cell, DTM_CELL * cell_to_merge, DTM *mnt, int color, int date, moving_state mstate)
{
	float	old_z_moyen; 
	int		new_nb_points;
	float	diff_moyen;

	if (cell->state == EMPTY) {
		cell->state = FILLED;
		cell->current_info.nb_points = cell_to_merge->current_info.nb_points;
		cell->current_info.firstDate = date;
		cell->current_info.lastDate = date;
		cell->current_info.z_moyen = cell_to_merge->current_info.z_moyen;
		cell->current_info.z_max = cell_to_merge->current_info.z_max;
		cell->current_info.z_min = cell_to_merge->current_info.z_min;
		cell->current_info.sigma_z = cell_to_merge->current_info.sigma_z;
		if (color != 0)
			cell->current_info.luminance = cell_to_merge->current_info.luminance;
		mnt->nb_cells ++;
	} else {
		old_z_moyen = cell->current_info.z_moyen;
		new_nb_points = cell->current_info.nb_points + cell_to_merge->current_info.nb_points;
		cell->current_info.z_moyen = (old_z_moyen * cell->current_info.nb_points +
				cell_to_merge->current_info.z_moyen * cell_to_merge->current_info.nb_points) / new_nb_points;

		cell->current_info.z_max = MAX(cell->current_info.z_max, cell_to_merge->current_info.z_max);
		cell->current_info.z_min = MIN(cell->current_info.z_min, cell_to_merge->current_info.z_min);

		diff_moyen = cell_to_merge->current_info.z_moyen - old_z_moyen;
		cell->current_info.sigma_z = ((cell->current_info.sigma_z * cell->current_info.sigma_z) * cell->current_info.nb_points + 
				(cell_to_merge->current_info.sigma_z * cell_to_merge->current_info.sigma_z) * cell_to_merge->current_info.nb_points +
				(diff_moyen * diff_moyen * cell->current_info.nb_points * cell_to_merge->current_info.nb_points) / new_nb_points) / 
			new_nb_points;

		cell->current_info.lastDate = date;
		if (color != 0) 
			cell->current_info.luminance = (cell->current_info.luminance * cell->current_info.nb_points +
					cell_to_merge->current_info.luminance * cell_to_merge->current_info.nb_points) / new_nb_points;
		cell->current_info.nb_points = new_nb_points;
	}

	cell->hstate = cell_to_merge->hstate;
	cell->mstate = mstate;
}

static bool
check_neighbour_cell(DTM* dtm, int i, int j)
{
	DTM_CELL* cell[4];
	cell[0] = dtm_one_cell(dtm, i-1, j);
	cell[1] = dtm_one_cell(dtm, i, j+1);
	cell[2] = dtm_one_cell(dtm, i, j-1);
	cell[3] = dtm_one_cell(dtm, i+1, j);

	int idx;
	int nb_non_visible = 0;
	int nb_visible_unknow = 0;
	int nb_visible_flat = 0;

	for (idx = 0; idx < 4; idx++) {
		if (cell[idx]->may_be_visible != SEEN_BOTH)
			nb_non_visible++;
		if (cell[idx]->may_be_visible == SEEN_BOTH && cell[idx]->hstate == UNKNOW)
			nb_visible_unknow++;
		if (cell[idx]->may_be_visible == SEEN_BOTH && cell[idx]->hstate == FLAT)
			nb_visible_flat++;
	}

	return (nb_non_visible > 0 && nb_visible_flat > 0);
}

/*
 * Merge de deux DTM possédant les mêmes caractéristiques (x,y, scale_* ) en
 * utilisant une méthode de moyennage
 * XXX : we probably can make merge smarter but we don't need a too smart merger ...
 */ 
static int 
dtm_merge(DTM*		mnt_to_merge,
 T3D*			RobotRed_2_Start,
 DTM*			mnt,
 int			color)
{
	int date;
	DTM_CELL	*cell_to_merge, *cell;
	int i,j;
	int shift_i, shift_j;
	int i_min, i_max, j_min, j_max;

	struct dynamic_fuse_data* data = mnt_to_merge->data;
	struct tab_min_max* limits = &data->tab;

	/* valididité des paramètres */
	if (mnt->cells == NIL_CELLS)
		return dtmError (ERR_DTM_CELLS_NON_ALLOUEES);

	if (mnt_to_merge->cells == NIL_CELLS)
		return dtmError (ERR_DTM_CELLS_NON_ALLOUEES);

	/* Slide the DTM */
	if (mnt->type == DTM_SLIDING) {
		dtm_slideDtm (mnt, RobotRed_2_Start);
	}

	if (mnt->scale_x != mnt_to_merge->scale_x ||
		mnt->scale_y != mnt_to_merge->scale_y)
		return dtmError (ERR_DTMS_INCOMPATIBLES);

	shift_i = mnt->i_origine - mnt_to_merge->i_origine;
	shift_j = mnt->j_origine - mnt_to_merge->j_origine;

	date = mnt->nb_im3d + 1;

	i_min = MAX (0, - shift_i);
	i_max = MIN (mnt_to_merge->nblig, mnt->nblig - shift_i);
	j_min = MAX (0, - shift_j);
	j_max = MIN (mnt_to_merge->nbcol, mnt->nbcol - shift_j);

	j_min = MAX(0, -shift_j);
	j_max = MIN(limits->size, (size_t)(mnt->nbcol - shift_j));

	for (j= j_min; j < j_max; j++)
	{
		int min = limits->tab[j].min;
		int max = limits->tab[j].max;
		if (max != -1) 
		{
			i_min = MAX(min, -shift_i);
			i_max = MIN(max, mnt->nblig - shift_i);
			for (i = i_min; i < i_max; i++) 
			{
				cell = dtm_one_cell (mnt, i+shift_i, j+shift_j);
				cell_to_merge = dtm_one_cell( mnt_to_merge, i, j);

				if (cell_to_merge->state == FILLED) {
					/*
					 * If the new cell and the last cell are in the same height
					 * state, we consider that the indication is confirmed, so we
					 * merge it 
					 *
					 * if the cell is empty and not in the unknow step, just merge
					 * the information
					 * */
					if ((cell->state == EMPTY && cell->hstate != UNKNOW) ||
							(cell->hstate == cell_to_merge->hstate)) {
						dtm_merge_cell(cell, cell_to_merge, mnt, color, date, STATIC);
					} else if (cell->hstate == FLAT) {
						/* The cell was flat, we have unknow or vertical case
						 * in vertical case, we consider an dynamic object
						 * in unknow case, the cell became unknow (but we keep floor information)
						 */
						memcpy(&cell->floor_info, &cell->current_info, sizeof(DTM_CELL_INFO));
						cell->state = EMPTY;
						mnt->nb_cells --;
						if (cell_to_merge->state == UNKNOW) 
							cell->hstate = UNKNOW;
						else
							dtm_merge_cell(cell, cell_to_merge, mnt, color, date, DYNAMIC);
					} else if (cell->hstate == VERTICAL) {
						/*
						 * The cell was VERTICAL, the cell to merge is flat or unknow
						 * consider something dynamic has moved from this cell 
						 * */
						if (cell_to_merge->state == UNKNOW) {
							memset(&cell->current_info, 0, sizeof(DTM_CELL_INFO));
							cell->hstate = UNKNOW;
							cell->state = EMPTY;
							mnt->nb_cells--;
						} else {
							memcpy(&cell->current_info, &cell->floor_info, sizeof(DTM_CELL_INFO));
							if (cell->current_info.nb_points == 0) {
								cell->state = EMPTY;
								mnt->nb_cells --;
							}
							dtm_merge_cell(cell, cell_to_merge, mnt, color, date, STATIC);
						}
					} else { /* cell->hstate == UNKNOW */
						if (cell_to_merge->state == FLAT) {
							/* Copy back the information of the floor, and merge the new information */
							memcpy(&cell->current_info, &cell->floor_info, sizeof(DTM_CELL_INFO));
							dtm_merge_cell(cell, cell_to_merge, mnt, color, date, STATIC);
						} else {
							/* Just integrate the vertical cell, floor_info already contains floor information */
							dtm_merge_cell(cell, cell_to_merge, mnt, color, date, DYNAMIC);
						}
					}
				} else if (dtm_visible_cell(cell_to_merge)) { 
					/* If the cell must be visible and we don't see it, it is
					 * probably because there is some obstacle on the ray vision.
					 * In this case, we can't say anything useful about it, but we can't just say that the zone is currently unknow
					 */
					if (cell->hstate == FLAT) {
						if (check_neighbour_cell(mnt_to_merge, i, j)) {
							continue;
						} else {
							/*
							 * zone was flat, it becomes unknow but we keep information about floor 
							 */
							memcpy(&cell->floor_info, &cell->current_info, sizeof(DTM_CELL_INFO));
							cell->state = EMPTY;
							mnt->nb_cells --;
							cell->hstate = UNKNOW;
						}
					} 

					if (cell->hstate == VERTICAL) {
						/* 
						 * zone was vertical, it becomes unknow 
						 */
						memset(&cell->current_info, 0, sizeof(DTM_CELL_INFO));
						cell->state = EMPTY;
						mnt->nb_cells --;
						cell->hstate = UNKNOW;
					}
				}
			}
		}
	}

	mnt->cells = CELLS_DETERMINEES;
	if (color != 0)
		mnt->luminance = LUMINANCES_DETERMINEES;
	mnt->nb_im3d++;

	return (DTM_OK);
}

DTM* 
prepare_dynamic_fuse(
		T3D* sensorLeftToMain, 
		T3D* sensorRightToMain,
		double *intrinsincParamsRectified,
		int nbLines, 
		int nbColumns,
		int reduction,
		double max_depth,
		double scale_x, 
		double scale_y)
{
	DTM* res = create_dtm();
	/*
	 * XXX
	 * It is not the right way to compute it
	 * We need to compute the size max of the non parallel side of the
	 * trapezoid, and compute size_x from this value
	 *
	 * It doesn't work for really strange reason, it leads to random crash in
	 * the dtm if we compute data before cells (and I'm really tired about dtm so ...)
	 *
	 * 3*max_depth is not correct, but maximise its value for the kind of
	 * camera we have
	 */
	double size_x = 3 * max_depth;
	double size_y = size_x;
	size_t i;

	if (res == NULL) 
		return res;

	init_dtm_cells(res, scale_x, scale_y, size_x / scale_x, size_y / scale_y,
			-size_x/2, -size_y/2, DTM_FIXED, X_LIGNES);

	res->data = malloc(sizeof(struct dynamic_fuse_data));
	if (res->data == NULL) {
		destroy_dtm(res);
		return NULL;
	}
	struct dynamic_fuse_data *data = (struct dynamic_fuse_data*) res->data;
	data->tab.tab = malloc(sizeof(struct min_max) * res->nbcol);
	data->tab.size = res->nbcol;

#ifndef APPLY_VISIBILITY
	/* 
	 * If we don't apply visibility, data->tab.tab is not computed, so we need
	 * to fill it by hand
	 */
	for (i = 0; i < data->tab.size; i++)
	{
		data->tab.tab[i].min = 0;
		data->tab.tab[i].max = res->nblig;
	}
#endif
	clear_dynamic_fuse_data(res, data);

	_compute_extremum_visible_cell(sensorRightToMain, intrinsincParamsRectified,
			nbLines, nbColumns, reduction, &data->ext, 1, 3, max_depth, false);
	_compute_extremum_visible_cell(sensorLeftToMain, intrinsincParamsRectified,
			nbLines, nbColumns, reduction, &data->ext, 0, 2, max_depth, true);

	// clear dtm, we don't need anymore the visibility image
	clear_dtm(res);

	return res;
}

void
clean_dynamic_fuse(DTM* dtm)
{
	struct dynamic_fuse_data *data = (struct dynamic_fuse_data*) dtm->data;
	if (data != NULL) {
		if (data->tab.tab != NULL) {
			free(data->tab.tab);
			data->tab.tab = NULL;
		}
		free(data);
		data = NULL;
	};

	destroy_dtm(dtm);
}




int
dtm_dynamic_fuse (DATA_IM3D*		im3d,
		  T3D*			RobotRed_2_Start,
		  T3D*			sensorToOrigin,
		  DTM*			mnt,
		  DTM*			tmp_mnt,
		  int			color)
{
	t3dConvertTo(T3D_EULER, sensorToOrigin);
	clear_dtm(tmp_mnt);
	clear_dynamic_fuse_data(tmp_mnt, tmp_mnt->data);
	set_dtm_center(tmp_mnt, sensorToOrigin->euler.euler[3], sensorToOrigin->euler.euler[4]);
	dtm_fuseData(im3d, RobotRed_2_Start, tmp_mnt, 0, color);
#ifdef APPLY_VISIBILITY
	dtm_apply_visibility_pattern(tmp_mnt, RobotRed_2_Start->euler.euler[3], 
										  RobotRed_2_Start->euler.euler[4],
										  sensorToOrigin->euler.euler[0]);
#endif

	evaluate_hstate(tmp_mnt, RobotRed_2_Start, ((struct dynamic_fuse_data*)tmp_mnt->data)->strategy);
	dtm_merge(tmp_mnt, RobotRed_2_Start, mnt, color);

	return 0;
}

dtm_dynamic_fuse_strategy dtm_dynamic_get_strategy(DTM* dtm)
{
	assert(dtm != NULL);
	assert(dtm->data != NULL);

	return ((struct dynamic_fuse_data*) dtm->data)->strategy;

}
void dtm_dynamic_set_strategy(DTM* dtm, dtm_dynamic_fuse_strategy strat)
{
	assert(dtm != NULL);
	assert(dtm->data != NULL);

	((struct dynamic_fuse_data*) dtm->data)->strategy = strat;
}
