/* Les includes génériques de C */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* structures, protos and errors of the lib */
#include "libdtm.h"

/**********************************************************************
 *
 * Save a DTM in an ASCII file.
 *
 **********************************************************************/

int
dtm_writeAsciiDtm (DTM*	mnt,
		   FILE*	file)

{
  int		i, j;		/* Pour boucler */
  DTM_CELL*	cell;

  /* write the general DTM info */
  fprintf (file, DTM_ASCII_FF);
  fprintf (file, "\n");
  fprintf (file, "%d %d %d %d %d\n", 
	   mnt->type, mnt->stockage, mnt->cells, mnt->bornes, mnt->luminance);
  fprintf (file, "%d\n", mnt->nb_im3d);
  fprintf (file, "%e %e\n", mnt->x_origine, mnt->y_origine);
  fprintf (file, "%d %d\n", mnt->i_origine, mnt->j_origine);
  fprintf (file, "%e %e\n", mnt->scale_x, mnt->scale_y);
  fprintf (file, "%e %e\n", mnt->size_x, mnt->size_y);
  fprintf (file, "%d %d\n", mnt->nblig, mnt->nbcol);
  fprintf (file, "%d\n", mnt->nb_cells);
  fprintf (file, "%d %d %d %d\n", mnt->i_min, mnt->i_max, mnt->j_min, mnt->j_max);
  if (mnt->bornes == BORNES_DTM_CALCULEES)
    fprintf (file, "%e %e\n", mnt->z_min, mnt->z_max);
  else
    fprintf (file, "%e %e\n", 0.0, 0.0);
  
  /* Write all the points */
  cell = mnt->cells_tab;
  for (i = 0; i < mnt->nblig; i++) {
    for (j = 0; j < mnt->nbcol; j++) {
      fprintf (file, "%d %d %d %d %d %d %d %f %f %f\n",
	       cell->state, cell->hstate, cell->may_be_visible, 
		   cell->current_info.nb_points, cell->current_info.luminance,
		   cell->current_info.firstDate, cell->current_info.lastDate,
	       cell->current_info.z_moyen, cell->current_info.sigma_z, 
		   cell->current_info.z_max);
      cell++;
    }
    fprintf (file, "\n");
  }

  return (DTM_OK);
}

/**********************************************************************
 *
 * Reads a DTM in an ASCII file.
 *
 **********************************************************************/

DTM*
dtm_readAsciiDtm (FILE*	file)

{
  int		i, j;		/* Pour boucler */
  char          keyword[256];
  DTM*		mnt;
  DTM_CELL*	cell;
  int		state, hstate, nb_points, luminance, firstDate, lastDate;
  int		may_be_visible;
  float		z_moyen, sigma_z, z_max;

  int  type_info, stockage_info, cells_infos, bornes_info, luminance_info;

  /* Creates the DTM */
  mnt = create_dtm ();
  if (mnt == NULL) {
    dtmError (ERR_DTM_MALLOC);
    return (NULL);
  }

  /* Read the general DTM info */
  fscanf (file, "%s", keyword);
  if (strcmp (keyword, DTM_ASCII_FF) != 0) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }

  if (fscanf (file, "%d %d %d %d %d\n", 
	      &type_info, &stockage_info, &cells_infos, 
	      &bornes_info, &luminance_info) != 5) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  mnt->type = type_info;
  mnt->stockage = stockage_info;
  mnt->cells = cells_infos;
  mnt->bornes = bornes_info;
  mnt->luminance = luminance_info; /* (we use these intermediate variables to
				      avoid a gcc warning */  
  if (fscanf (file, "%d\n", &(mnt->nb_im3d)) != 1) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%le %le\n", &(mnt->x_origine), &(mnt->y_origine)) != 2) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%d %d\n", &(mnt->i_origine), &(mnt->j_origine)) != 2) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%le %le\n", &(mnt->scale_x), &(mnt->scale_y)) != 2) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%le %le\n", &(mnt->size_x), &(mnt->size_y)) != 2) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%d %d\n", &(mnt->nblig), &(mnt->nbcol)) != 2) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%d\n", &(mnt->nb_cells)) != 1) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%d %d %d %d\n", 
	      &(mnt->i_min), &(mnt->i_max), &(mnt->j_min), &(mnt->j_max)) != 4) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%le %le\n", &(mnt->z_min), &(mnt->z_max)) != 2) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }

  /* Allocation of the DTM cell array */
  mnt->cells = NIL_CELLS;	/* In order to not get an error from the
				   init_dtm_cells fct  */

  switch (mnt->stockage) {
  case X_LIGNES:
    if (init_dtm_cells (mnt, mnt->scale_x, mnt->scale_y, mnt->nblig, mnt->nbcol,
			mnt->x_origine, mnt->y_origine, mnt->type, mnt->stockage) != DTM_OK) {
      dtmError (ERR_DTM_MALLOC);
      return (NULL);
    }
    break;
  case X_COLONNES:
    if (init_dtm_cells (mnt, mnt->scale_x, mnt->scale_y, mnt->nbcol, mnt->nblig,
			mnt->x_origine, mnt->y_origine, mnt->type, mnt->stockage) != DTM_OK) {
      dtmError (ERR_DTM_MALLOC);
      return (NULL);
    }
    break;
  default:
    dtmError (ERR_DTM_CAS_IMPOSSIBLE);
    return (NULL);
    /*     break; */
    /* Unreached */
  }
  
  /* Read all the points */
  cell = mnt->cells_tab;
  for (i = 0; i < mnt->nblig; i++) {
    for (j = 0; j < mnt->nbcol; j++) {
      if (fscanf (file, "%d %d %d %d %d %d %d %f %f %f\n",
		  &state, &hstate, &may_be_visible, &nb_points, &luminance, 
		  &firstDate, &lastDate, &z_moyen, &sigma_z, &z_max) != 10) {
	dtmError (ERR_DTM_FILEFORMAT);
	return (NULL);
      }
      cell->state = state;
	  cell->hstate = hstate;
	  cell->may_be_visible = may_be_visible;
      cell->current_info.nb_points = nb_points;
      cell->current_info.luminance = luminance;
      cell->current_info.firstDate = firstDate;
      cell->current_info.lastDate = lastDate;
      cell->current_info.z_moyen = z_moyen;
      cell->current_info.sigma_z = sigma_z;
	  cell->current_info.z_max = z_max;
      cell++;
    }
    fscanf (file, "\n");
  }
  
  /* Update some flags */
  mnt->cells = CELLS_DETERMINEES;

  return (mnt);
}


/**********************************************************************
 *
 * Save a DTM in an BINARY file.
 *
 **********************************************************************/

int
dtm_writeBinaryDtm (DTM*	mnt,
		FILE*	file)

{
  int		i;
  unsigned int		nbCells;

  /* write the general DTM info (in ASCII) */
  fprintf (file, DTM_ASCII_FF);
  fprintf (file, "\n");
  fprintf (file, "%d %d %d %d %d\n", 
	   mnt->type, mnt->stockage, mnt->cells, mnt->bornes, mnt->luminance);
  fprintf (file, "%d\n", mnt->nb_im3d);
  fprintf (file, "%e %e\n", mnt->x_origine, mnt->y_origine);
  fprintf (file, "%d %d\n", mnt->i_origine, mnt->j_origine);
  fprintf (file, "%e %e\n", mnt->scale_x, mnt->scale_y);
  fprintf (file, "%e %e\n", mnt->size_x, mnt->size_y);
  fprintf (file, "%d %d\n", mnt->nblig, mnt->nbcol);
  fprintf (file, "%d\n", mnt->nb_cells);
  fprintf (file, "%d %d %d %d\n", mnt->i_min, mnt->i_max, mnt->j_min, mnt->j_max);
  fprintf (file, "%e %e\n", mnt->z_min, mnt->z_max);

  /* Write all the points */
  nbCells = mnt->nblig * mnt->nbcol;
  if (fwrite (mnt->cells_tab, sizeof (DTM_CELL), nbCells, file) != nbCells)
    return dtmError (ERR_DTM_WRITE_FILE);
  
  return (DTM_OK);
}

/**********************************************************************
 *
 * Reads a DTM in an BINARY file.
 *
 **********************************************************************/

DTM*
dtm_readBinaryDtm (FILE * file)
{
  int		i;		/* Pour boucler */
  char          keyword[256];
  DTM*		mnt;
  unsigned int		nbCells;

  int  type_info, stockage_info, cells_infos, bornes_info, luminance_info;

  /* Creates the DTM */
  mnt = create_dtm ();
  if (mnt == NULL) {
    dtmError (ERR_DTM_MALLOC);
    return (NULL);
  }

  /* Read the general DTM info */
  fscanf (file, "%s", keyword);
  if (strcmp (keyword, DTM_ASCII_FF) != 0) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }

  if (fscanf (file, "%d %d %d %d %d\n", 
	      &type_info, &stockage_info, &cells_infos, 
	      &bornes_info, &luminance_info) != 5) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  mnt->type = type_info;
  mnt->stockage = stockage_info;
  mnt->cells = cells_infos;
  mnt->bornes = bornes_info;
  mnt->luminance = luminance_info; /* (we use these intermediate variables to
				      avoid a gcc warning */  

  if (fscanf (file, "%d\n", &(mnt->nb_im3d)) != 1) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%le %le\n", &(mnt->x_origine), &(mnt->y_origine)) != 2) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%d %d\n", &(mnt->i_origine), &(mnt->j_origine)) != 2) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%le %le\n", &(mnt->scale_x), &(mnt->scale_y)) != 2) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%le %le\n", &(mnt->size_x), &(mnt->size_y)) != 2) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%d %d\n", &(mnt->nblig), &(mnt->nbcol)) != 2) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%d\n", &(mnt->nb_cells)) != 1) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%d %d %d %d\n", 
	      &(mnt->i_min), &(mnt->i_max), &(mnt->j_min), &(mnt->j_max)) != 4) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "%le %le\n", &(mnt->z_min), &(mnt->z_max)) != 2) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }

  /* Allocation of the DTM cell array */
  mnt->cells = NIL_CELLS;	/* In order to not get an error from the
				   init_dtm_cells fct  */

  switch (mnt->stockage) {
  case X_LIGNES:
    if (init_dtm_cells (mnt, mnt->scale_x, mnt->scale_y, mnt->nblig, mnt->nbcol,
			mnt->x_origine, mnt->y_origine, mnt->type, mnt->stockage) != DTM_OK) {
      dtmError (ERR_DTM_MALLOC);
      return (NULL);
    }
    break;
  case X_COLONNES:
    if (init_dtm_cells (mnt, mnt->scale_x, mnt->scale_y, mnt->nbcol, mnt->nblig,
			mnt->x_origine, mnt->y_origine, mnt->type, mnt->stockage) != DTM_OK) {
      dtmError (ERR_DTM_MALLOC);
      return (NULL);
    }
    break;
  default:
    dtmError (ERR_DTM_CAS_IMPOSSIBLE);
    return (NULL);
    /*     break; */
    /* Unreached */
  }
  
  /* Read all the points */
  nbCells = mnt->nblig * mnt->nbcol;
  if (fread (mnt->cells_tab, sizeof (DTM_CELL), nbCells, file) != nbCells) {
    dtmError (ERR_DTM_READ_FILE);
    return (NULL);
  }

  /* Update some flags */
  mnt->cells = CELLS_DETERMINEES;

  return (mnt);
}

