/* Les includes génériques de C */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* Library includes */
#include "libdtm.h"

/**********************************************************************
 *
 * Moves the DTM so that the current robot position is at the center of the
 * DTM. 
 *
 **********************************************************************/

int
dtm_slideDtm (DTM*		dtm,
	      T3D*		RobotRed_2_Start)

{
  DTM*	dtm2;			/* The DTM used as a buffer */

  int	i_robot, j_robot;	/* Robot in integer coordinates */
  int	i0, j0;			/* Integer coordinates of the world reference
				   frame */ 
  int	nb_pix_metre_i;
  int	nb_pix_metre_j;		/* Number of dtm cells per meter */

  int   shift_i, shift_j;	/* Translation to apply on the DTM  */
  int   iMin, iMax;
  int   jMin, jMax;
  int   i, j;
  DTM_CELL* cell;
  DTM_CELL* cell2;


  /* Various initializations */
  t3dConvertTo (T3D_EULER, RobotRed_2_Start);
  i0 = dtm->i_origine;
  j0 = dtm->j_origine;

  /* Compute the integer coordinates of the current robot position with respect
     to the current DTM */
  switch (dtm->stockage) {
  case X_LIGNES:
    nb_pix_metre_i = lround (1.0 / dtm->scale_x);
    nb_pix_metre_j = lround (1.0 / dtm->scale_y);
    i_robot = nb_pix_metre_i * RobotRed_2_Start->euler.euler[3] + i0;
    j_robot = nb_pix_metre_j * RobotRed_2_Start->euler.euler[4] + j0;
    break;
  case X_COLONNES:
    return dtmError (ERR_DTM_FCT_NON_IMPLEMENTEE);
    /*     break; */
    /* Unreached */
  default:
    /*     break; */
    /*     Unreached */
    return dtmError (ERR_DTM_CAS_IMPOSSIBLE);
  }

  shift_i = i_robot - (dtm->nblig / 2);
  shift_j = j_robot - (dtm->nbcol / 2);

  /* Only effectivly slide the dtm if we have a sensible translation to do */
  if ((abs(shift_i) < (dtm->nblig / 20)) && (abs(shift_j) < dtm->nbcol / 20))
	  return (DTM_OK);

  printf ("Shift_i = %d; shift_j = %d\n", shift_i, shift_j);

  /* Creates the new DTM */
  dtm2 = create_dtm ();
  if (dtm2 == NULL)
    return dtmError (ERR_DTM_MALLOC);

  /* Initialisation */
  if (init_dtm_cells (dtm2, dtm->scale_x, dtm->scale_y, dtm->nblig, dtm->nbcol, 
		      dtm->x_origine + (shift_i *  dtm->scale_x),
		      dtm->y_origine + (shift_j * dtm->scale_y), 
		      dtm->type, dtm->stockage) != DTM_OK) 
    return dtmError (ERR_DTM_MALLOC);

  /* Now loop on the new DTM cells, and copy the ones form the old one */
  iMin = MAX (0, - shift_i);
  iMax = MIN (dtm2->nblig, dtm2->nblig - shift_i);
  jMin = MAX (0, - shift_j);
  jMax = MIN (dtm2->nbcol, dtm2->nbcol - shift_j);

  printf ("I: %d to %d ; J: %d to %d\n", iMin, iMax, jMin, jMax);
  for (i = iMin; i < iMax; i++) {
    for (j = jMin; j < jMax; j++) {
      cell2 = dtm2->cells_tab + (i * dtm2->nbcol) + j;
      cell = dtm->cells_tab + ((i + shift_i) * dtm->nbcol) + (j + shift_j);
      memcpy (cell2, cell, sizeof(DTM_CELL));
    }
  }

  /* Copy the header informations and the cells, update the min/max informations
     */ 
  dtm->x_origine = dtm2->x_origine;
  dtm->y_origine = dtm2->y_origine;
  dtm->i_origine = dtm2->i_origine;
  dtm->j_origine = dtm2->j_origine;
  memcpy (dtm->cells_tab, dtm2->cells_tab, 
	  dtm->nblig * dtm->nbcol * sizeof (DTM_CELL));
  if (dtm->bornes == BORNES_DTM_CALCULEES) 
    dtm_minmax (dtm);

  destroy_dtm (dtm2);

  return (DTM_OK);
}

