/* Les includes génériques de C */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Les structures, protos et erreurs de la lib */
#include "libdtm.h"

/**********************************************************************
 *
 * Get the bounds on various attributes of the dtm cells
 *
 **********************************************************************/

int
dtm_minmax (DTM*     mnt)

{
  int		i, j;
  DTM_CELL*	cell;
  int		nblig;
  int		nbcol;

  nblig = mnt->nblig;
  nbcol = mnt->nbcol;

  cell = mnt->cells_tab;
  for (i = 0; i < nblig; i++) {
    for (j = 0; j < nbcol; j++) {
      if (cell->state == FILLED) {
	mnt->i_max = MAX (mnt->i_max, i);
	mnt->i_min = MIN (mnt->i_min, i);

	mnt->j_max = MAX (mnt->j_max, j);
	mnt->j_min = MIN (mnt->j_min, j);
	
	mnt->z_max = MAX (mnt->z_max, cell->current_info.z_moyen);
	mnt->z_min = MIN (mnt->z_min, cell->current_info.z_moyen);

	mnt->sigmaMax = MAX (mnt->sigmaMax, cell->current_info.sigma_z);
	mnt->sigmaMin = MIN (mnt->sigmaMin, cell->current_info.sigma_z);

	mnt->nbPtsMax = MAX (mnt->nbPtsMax, cell->current_info.nb_points);
	mnt->nbPtsMin = MIN (mnt->nbPtsMin, cell->current_info.nb_points);
      }
      cell++;
    }
  }
  
  if (mnt->luminance == LUMINANCES_DETERMINEES) {
    cell = mnt->cells_tab;
    for (i = 0; i < nblig; i++) {
      for (j = 0; j < nbcol; j++) {
	if (cell->state == FILLED) {
	  mnt->colorMin = MIN (mnt->colorMin, cell->current_info.luminance);
	  mnt->colorMax = MAX (mnt->colorMax, cell->current_info.luminance);
	}
	cell++;
      }
    }
  }

  mnt->bornes = BORNES_DTM_CALCULEES;

  return (DTM_OK);
}

/**********************************************************************
 *
 * Returns the elevation of a point specified in real coordinates
 *
 **********************************************************************/

int
dtm_getElevation (DTM*		mnt,
		  double	x,
		  double	y,
		  double*	z)

{
  int		i, j;

  i = (int) (floor ((x - mnt->x_origine) / mnt->scale_x));
  j = (int) (floor ((y - mnt->y_origine) / mnt->scale_y));

  if ((i < 0) || (i > mnt->nblig) || (j < 0) || (j > mnt->nbcol))
    return dtmError (ERR_DTM_OUT_OF_DTM);

  *z = (mnt->cells_tab + (i * mnt->nbcol) + j)->current_info.z_moyen;

  return (DTM_OK);
}


/**********************************************************************
 *
 * 
 *
 **********************************************************************/

int
dtm_boundLastDate (DTM*	mnt,
		   int	maxLastDate)

{
  int		i, j;
  DTM_CELL*	cell;
  int		nblig;
  int		nbcol;

  nblig = mnt->nblig;
  nbcol = mnt->nbcol;

  cell = mnt->cells_tab;
  for (i = 0; i < nblig; i++) {
    for (j = 0; j < nbcol; j++) {
      if ((cell->state == FILLED) && (cell->current_info.lastDate >= maxLastDate)) 
	cell->state = EMPTY;
      cell++;
    }
  }
  
  return (DTM_OK);
}

/**********************************************************************
 *
 * 
 *
 **********************************************************************/

int
dtm_boundFirstDate (DTM*	mnt,
		    int		maxFirstDate)

{
  int		i, j;
  DTM_CELL*	cell;
  int		nblig;
  int		nbcol;

  nblig = mnt->nblig;
  nbcol = mnt->nbcol;

  cell = mnt->cells_tab;
  for (i = 0; i < nblig; i++) {
    for (j = 0; j < nbcol; j++) {
      if ((cell->state == FILLED) && (cell->current_info.firstDate >= maxFirstDate)) 
	cell->state = EMPTY;
      cell++;
    }
  }
  
  return (DTM_OK);
}



/**********************************************************************
 *
 * Construction d'un DTM à partir d'une ima8
 *
 **********************************************************************/

DTM*
ima8_to_dtm (double*	q_image,
	     int	nblig,
	     int	nbcol,
	     double	scale_x,
	     double	scale_y,
	     int	stockage)

{
  int		i, j;
  DTM*		mnt;
  DTM_CELL*	cell;
  double	pixel;
  double	min;
  int		nb_cells = 0;
  double*	q_image_ptr;	/* Pour rechercher la valeur min */

  /* On crée le mnt */
  mnt = create_dtm ();
  if (mnt == NULL) {
    dtmError (ERR_DTM_MALLOC);
    return (NULL);
  }
  
  switch (stockage) {
  case X_LIGNES:
    if (init_dtm_cells (mnt, scale_x, scale_y, 
			nblig, nbcol,
			0, 0, DTM_FIXED, stockage) != DTM_OK)
      return (NULL);
    break;
  case X_COLONNES:
    if (init_dtm_cells (mnt, scale_x, scale_y, 
			nbcol, nblig,
			0, 0, DTM_FIXED, stockage) != DTM_OK)
      return (NULL);
    break;
  default:
    dtmError (ERR_DTM_CAS_IMPOSSIBLE);
    return (NULL);
/*     unreached */
/*     break; */
  }

  /* On détermine la valeur min de l'ima8 (celle pour laquelle on n'a pas
     d'informations, par définition) */
  q_image_ptr = q_image;
  min = *q_image_ptr;
  for (i = 0; i < nblig; i++) {
    for (j = 0; j < nbcol; j++) {
      min = MIN (*q_image, min);
      q_image_ptr++;
    }
  }

  /* Maintenant on va boucler sur le MNT et copier les points correspondants
     de l'ima8 */
  cell = mnt->cells_tab;
  for (i = 0; i < mnt->nblig; i++) {
    for (j = 0; j < mnt->nbcol; j++) {
      /* On a inversé i et j pour avoir la même orientation entre les
	 DTM construits et les DTMS lus par rd_inrimage2 TODO !!! : changer
	 X_ligne et X_colonnes n'aurait pas le me6me effet ??? Je ne sais même
	 plus... */
      pixel = *(q_image + (j * nblig) + i);
      if (pixel == min) 
	/* Le point n'est pas valide */
	cell->state = EMPTY;
      else {
	cell->state = FILLED;
	cell->current_info.z_moyen = MM_TO_M (pixel);
	cell->current_info.sigma_z = 0;
	cell->current_info.nb_points = 1;
	nb_cells++;
      }
      cell++;
    }
  }
  
  /* On met à jour quelques autres champs du nouveau MNT */
  mnt->cells = CELLS_DETERMINEES;
  mnt->nb_im3d = 1;
  mnt->nb_cells = nb_cells;

  return (mnt);
}

/**********************************************************************
 *
 * À partir d'un DTM rempli, en construit un plus petit qui encadre juste les
 * points effectivement perçus
 *
 **********************************************************************/

DTM*
dtm_autocrop (DTM*     mnt)

{
  DTM*		new_mnt;
  DTM_CELL*	new_cell;

  int		nblig;
  int		nbcol;

  int		i_min;
  int		j_min;
  int		new_nblig;
  int		new_nbcol;
  double	x_origine;
  double	y_origine;

  int		i, j;

  nblig = mnt->nblig;
  nbcol = mnt->nbcol;

  /* On détermine les bornes du grand MNT... */
  dtm_minmax (mnt);

  /* ... et différents paramètres qui déterminent la taille du nouveau. */
  i_min = mnt->i_min;
  j_min = mnt->j_min;
  new_nblig = mnt->i_max - mnt->i_min + 1;
  new_nbcol = mnt->j_max - mnt->j_min + 1;
  switch (mnt->stockage) {
  case X_LIGNES:
    x_origine = (i_min * mnt->scale_x) + mnt->x_origine;
    y_origine = (j_min * mnt->scale_y) + mnt->y_origine;
    break;
  case X_COLONNES:
    x_origine = (j_min * mnt->scale_x) + mnt->x_origine;
    y_origine = (i_min * mnt->scale_y) + mnt->y_origine;
    break;
  default:
    dtmError (ERR_DTM_CAS_IMPOSSIBLE);
    return (NULL);
/*     Unreached */
    /*     break; */
  }

  /* On crée le nouveau mnt */
  new_mnt = create_dtm ();
  if (new_mnt == NULL) {
    dtmError (ERR_DTM_MALLOC);
    return (NULL);
  }

  if (init_dtm_cells (new_mnt, mnt->scale_x, mnt->scale_y, 
		      new_nblig, new_nbcol, 
		      x_origine, y_origine, 
		      mnt->type, mnt->stockage) != DTM_OK)
    return (NULL);

  /* Maintenant on va boucler sur le nouveau MNT et copier les points
     correspondants de l'ancien */
  new_cell = new_mnt->cells_tab;
  for (i = 0; i < new_mnt->nblig; i++) {
    for (j = 0; j < new_mnt->nbcol; j++) {
      memcpy ((DTM_CELL*) new_cell, 
	      (DTM_CELL*) (mnt->cells_tab + nbcol * (i + i_min) + (j + j_min)),
	      sizeof (DTM_CELL));
      new_cell++;
    }
  }

  /* On met à jour quelques autres champs du nouveau MNT */
  new_mnt->stockage = mnt->stockage;
  new_mnt->cells = mnt->cells;
  new_mnt->nb_im3d = mnt->nb_im3d;
  new_mnt->nb_cells = mnt->nb_cells;

  new_mnt->luminance = mnt->luminance;


/*   new_mnt->bornes = mnt->bornes; */

  return (new_mnt);
}

/**********************************************************************
 *
 * Crops a dtm
 *
 **********************************************************************/

DTM*
dtm_crop (DTM*     mnt,
	  int      iMin,
	  int      iMax,
	  int      jMin,
	  int      jMax)

{
  DTM*		newMnt;
  DTM_CELL*	new_cell;

  int		nblig;
  int		nbcol;

  int		new_nblig;
  int		new_nbcol;
  double	x_origine;
  double	y_origine;

  int		i, j = 0;

  nblig = mnt->nblig;
  nbcol = mnt->nbcol;

  /* Bounds the crop size to the bounds of the DTM */
  if (iMin < 0)
    iMin = 0;
  if (iMax >= mnt->nblig)
    iMax = mnt->nblig - 1;
  if (jMin < 0)
    jMin = 0;
  if (jMax >= mnt->nbcol)
    jMax = mnt->nbcol - 1;

  /* Sets the parameters required by the intialisation of the cropped dtm */
  new_nblig = iMax - iMin + 1;
  new_nbcol = jMax - jMin + 1;
  switch (mnt->stockage) {
  case X_LIGNES:
    x_origine = (iMin * mnt->scale_x) + mnt->x_origine;
    y_origine = (jMin * mnt->scale_y) + mnt->y_origine;
    break;
  case X_COLONNES:
    x_origine = (jMin * mnt->scale_x) + mnt->x_origine;
    y_origine = (iMin * mnt->scale_y) + mnt->y_origine;
    break;
  default:
    dtmError (ERR_DTM_CAS_IMPOSSIBLE);
    return (NULL);
/*     Unreached */
    /*     break; */
  }

  /* Creates the cropped dtm */
  newMnt = create_dtm ();
  if (newMnt == NULL) {
    dtmError (ERR_DTM_MALLOC);
    return (NULL);
  }

  if (init_dtm_cells (newMnt, mnt->scale_x, mnt->scale_y, 
		      new_nblig, new_nbcol, 
		      x_origine, y_origine, 
		      mnt->type, mnt->stockage) != DTM_OK)
    return (NULL);

  /* Loop in the cropped DTM and copy the cells */
  new_cell = newMnt->cells_tab;
  for (i = 0; i < newMnt->nblig; i++) {
    for (j = 0; j < newMnt->nbcol; j++) {
      memcpy ((DTM_CELL*) new_cell, 
	      (DTM_CELL*) (mnt->cells_tab + nbcol * (i + iMin) + (j + jMin)),
	      sizeof (DTM_CELL));
      new_cell++;
    }
  }

  /* update some fields of the cropped dtm */
  newMnt->stockage = mnt->stockage;
  newMnt->cells = mnt->cells;
  newMnt->nb_im3d = mnt->nb_im3d;
  newMnt->nb_cells = mnt->nb_cells;
  newMnt->luminance = mnt->luminance;

  /* Updates the geodesic position of its origin (if needed) */
  if ((mnt->geoReference == GEO_REFERENCE_SET)
      && (mnt->geoOrientation != NO_GEO_ORIENTATION)) {
    DTM_CELL* cell;
    double xCell, yCell;

    cell = newMnt->cells_tab;
    xCell = iMin * mnt->scale_x;
    yCell = jMin * mnt->scale_y;

    switch (mnt->geoOrientation) {
    case GEO_NORTH:
      newMnt->geodesicOrigin.easting = - yCell + mnt->geodesicOrigin.easting;
      newMnt->geodesicOrigin.northing = xCell + mnt->geodesicOrigin.northing;
      break;
    case GEO_WEST:
      newMnt->geodesicOrigin.easting = - xCell + mnt->geodesicOrigin.easting;
      newMnt->geodesicOrigin.northing = - yCell + mnt->geodesicOrigin.northing;
      break;
    case GEO_SOUTH:
      newMnt->geodesicOrigin.easting = yCell + mnt->geodesicOrigin.easting;
      newMnt->geodesicOrigin.northing = - xCell + mnt->geodesicOrigin.northing;
      break;
    case GEO_EAST:
      newMnt->geodesicOrigin.easting = xCell + mnt->geodesicOrigin.easting;
      newMnt->geodesicOrigin.northing = yCell + mnt->geodesicOrigin.northing;
      break;
    default:
      printf ("Nowhere !\n");
      dtmError(ERR_DTM_CAS_IMPOSSIBLE);
      return (NULL);
      /*     break; */
      /* Unreached */
    }
    newMnt->geodesicOrigin.alt = cell->current_info.z_moyen;
    newMnt->geodesicOrigin.nextEastMeridian = 
      mnt->geodesicOrigin.nextEastMeridian;
    dtm_wgs84utm2lla (newMnt->geodesicOrigin.easting, 
		      newMnt->geodesicOrigin.northing, 
		      newMnt->geodesicOrigin.nextEastMeridian,
		      &(newMnt->geodesicOrigin.lat),
		      &(newMnt->geodesicOrigin.lon));
    newMnt->geoReference = mnt->geoReference;
    newMnt->geoOrientation = mnt->geoOrientation;
  }

  return (newMnt);
}


/**********************************************************************
 *
 * À partir d'un DTM rempli, en construit un plus petit en le
 * sous-echantillonnant 
 *
 **********************************************************************/

DTM*
down_sample_dtm (DTM*	mnt,
		 int	reduction,
		 int	moyennage)

{
  DTM*		new_mnt;
  DTM_CELL*	new_cell;	/* Nouveau mnt et une de ses cellules */

  DTM_CELL*	cell;

  int		nblig;
  int		nbcol;
  int		new_nblig;
  int		new_nbcol;

  int		i, j, k, l;

  /* Validité des paramètres */
  if (reduction <= 0) {
    dtmError (ERR_DTM_TAUX_REDUCTION);
    return (NULL);
  }

  nblig = mnt->nblig;
  nbcol = mnt->nbcol;

  new_nblig = (nblig + 1) / 2;
  new_nbcol = (nbcol + 1) / 2;

  /* On crée le nouveau mnt */
  new_mnt = create_dtm ();
  if (new_mnt == NULL) {
    dtmError (ERR_DTM_MALLOC);
    return (NULL);
  }

  if (init_dtm_cells (new_mnt, 
		      reduction * mnt->scale_x, reduction * mnt->scale_y, 
		      new_nblig, new_nbcol,
		      mnt->x_origine, mnt->y_origine,
		      mnt->type, mnt->stockage) != DTM_OK)
    return (NULL);
  
  /* Maintenant on va boucler sur le nouveau MNT et copier les points
     correspondants de l'ancien */
  new_cell = new_mnt->cells_tab;
  new_mnt->nb_cells = 0;

  if (moyennage == 0) {		/* On fait un simple sous-echantillonnage
				   (c'est pas très bô). */
    for (i = 0; i < new_nblig; i++) {
      for (j = 0; j < new_nbcol; j++) {
	memcpy ((DTM_CELL*) new_cell, 
		(DTM_CELL*) (mnt->cells_tab 
			     + (nbcol * reduction * i) + (reduction * j)),
		sizeof (DTM_CELL));

	if (new_cell->state == FILLED)
	  new_mnt->nb_cells += 1;

	new_cell++;
      }
    }
  }
  else {			/* On fait un moyennage (c'est un peu mieux) */
    for (i = 0; i < new_nblig; i++) {
      for (j = 0; j < new_nbcol; j++) {
	/* On boucle sur le "pavé" qu'on va moyenner */
	for (k = 0; k < reduction; k++) {
	  for (l = 0; l < reduction; l++) {
	    cell =  mnt->cells_tab +
		    (nbcol * ((reduction * i) + k)) + ((reduction * j) + l);
	    if ((new_cell->current_info.nb_points + cell->current_info.nb_points) != 0) {
	      new_cell->current_info.z_moyen = ((new_cell->current_info.nb_points * new_cell->current_info.z_moyen)
				   + (cell->current_info.nb_points * cell->current_info.z_moyen))
				  / (new_cell->current_info.nb_points + cell->current_info.nb_points);
	      new_cell->state = FILLED;
	    }
	    new_cell->current_info.nb_points = (new_cell->current_info.nb_points + cell->current_info.nb_points);
	  }
	}

	if (new_cell->state == FILLED)
	  new_mnt->nb_cells += 1;
	
	new_cell++;
      }
    }
  }

  /* On met à jour quelques autres champs du nouveau MNT */
  new_mnt->cells = CELLS_DETERMINEES;
  new_mnt->nb_im3d = mnt->nb_im3d;

  return (new_mnt);
}


/**********************************************************************
 *
 * Mise à '0' d'une cellule d'un DTM. On fait ça à cause du fichier Geroms
 * original, qui a une cellule à une hauteur aberrante.
 *
 **********************************************************************/

int
void_dtm_cell (DTM*		mnt,
	       int		i,
	       int		j)
     

{
  DTM_CELL*	cell;

  if ((i < 0) || (i >= mnt->nblig) || (j < 0) || (j >= mnt->nbcol)) {
    printf ("Entre des valeurs 0 <= I < %d et 0 <= J < %d !\n",
	    mnt->nblig, mnt->nbcol);
    return dtmError(ERR_DTM_INDICES);
  }

  cell =  mnt->cells_tab + (mnt->nbcol * i) + j;
  cell->state = EMPTY;
  cell->current_info.nb_points = 0;
  cell->current_info.luminance = 0;
  cell->current_info.z_moyen = 0;
  cell->current_info.sigma_z = 0;

  return (DTM_OK);
}

/**********************************************************************
 *
 * Returns the nb of cells updated since the "date" given in parameter (date
 * in the sense of the number of images merged within the DTM)
 *
 **********************************************************************/

int
dtm_getUpdateSize (DTM*	mnt,
		   int	lastDate,
		   int* nbUpdatedCells)

{
  int		i, j;
  DTM_CELL*	cell;
  int		nblig;
  int		nbcol;

  int		iMin, iMax, jMin, jMax;

  nblig = mnt->nblig;
  nbcol = mnt->nbcol;

  *nbUpdatedCells = 0;
  iMin = nblig;
  iMax = 0;
  jMin = nbcol;
  jMax = 0;

  if (lastDate >= mnt->nb_im3d) 
    return dtmError (ERR_DTM_BAD_PARAMS);

  cell = mnt->cells_tab;
  for (i = 0; i < nblig; i++) {
    for (j = 0; j < nbcol; j++) {
      if ((cell->state == FILLED) && (cell->current_info.lastDate >= lastDate)) 
	*nbUpdatedCells += 1;
      cell++;
    }
  }
  
  return (DTM_OK);
}
